@@ -1,9 +1,9 @@
 /*
- Copyright (c) 2008 TrueCrypt Foundation. All rights reserved.
+ Copyright (c) 2008 TrueCrypt Developers Association. All rights reserved.
 
- Governed by the TrueCrypt License 2.4 the full text of which is contained
- in the file License.txt included in TrueCrypt binary and source code
- distribution packages.
+ Governed by the TrueCrypt License 3.0 the full text of which is contained in
+ the file License.txt included in TrueCrypt binary and source code distribution
+ packages.
 */
 
 #ifndef TC_HEADER_Core_Unix_MountedFilesystem
@@ -18,6 +18,7 @@ namespace TrueCrypt
 	public:
 		DevicePath Device;
 		DirectoryPath MountPoint;
+		string Type;
 	};
 
 	typedef list < shared_ptr <MountedFilesystem> > MountedFilesystemList;
